from asyncio_rpc.models import RPCBase
from abc import ABC, abstractmethod


class AbstractRPCCommLayer(ABC):
    """
    Abstract baseclass for RPC communication layer
    """
    @abstractmethod
    async def publish(self, rpc_instance: RPCBase):
        """
        Publish a RPCBase subclass to the other end,
        either a RPCServer or RPCClient.
        """

    @abstractmethod
    async def do_subscribe(self):
        """
        Initialize subscription for receiving messages.
        This is needed for Redis to make sure all
        message are received, this might not be the case
        for other communication layers
        """

    @abstractmethod
    async def subscribe(self, on_rpc_event_callback: callable):
        """
        Subscribe and listen for messages. The on_rpc_event_callback
        function is called on every received RPCBase message
        """

    @abstractmethod
    async def unsubscribe(self):
        """
        Stop subscription for receiving messages.
        This is needed for Redis and might not be needed
        for other communication layers
        """



import pytest

from tests.utils import (
    rpc_client as rpc_client_fixture
)

from asyncio_rpc.models import RPCMessage

rpc_client = rpc_client_fixture


async def stop_rpc_client_on_rpc_message(rpc_client, expected_rpc_message):
    async def on_rpc_message(rpc_message: RPCMessage, channel):
        assert isinstance(rpc_message, RPCMessage)

        await rpc_client.queue.put(b'END')
        await rpc_client.rpc_commlayer.unsubscribe()

        assert rpc_message == expected_rpc_message

    return on_rpc_message


@pytest.mark.asyncio
async def test_rpc_message(rpc_client):
    await rpc_client.rpc_commlayer.do_subscribe()

    rpc_message = RPCMessage(
        uid='1', namespace='TEST', data={'foo': 'bar'})

    # publish to self
    await rpc_client.rpc_commlayer.publish(
        rpc_message, channel=rpc_client.rpc_commlayer.subchannel)

    await rpc_client.serve(
        on_rpc_message=await stop_rpc_client_on_rpc_message(
            rpc_client, rpc_message))

    await rpc_client.rpc_commlayer.close()



import pytest
import asyncio

from os import environ
from uuid import uuid4


from asyncio_rpc.serialization import msgpack as msgpack_serialization

from asyncio_rpc.server import RPCServer
from asyncio_rpc.client import RPCClient
from asyncio_rpc.models import RPCCall, RPCStack
from asyncio_rpc.commlayers.redis import RPCRedisCommLayer

# Set to env_var REDIS_HOST or 'localhost' as default
REDIS_HOST = environ.get('REDIS_HOST', 'localhost')


async def rpc_commlayer(subchannel, pubchannel, host=REDIS_HOST):
    """
    Get a RPCRedisCommLayer with subchannel/pubchannel
    """
    return await RPCRedisCommLayer.create(
            subchannel=subchannel, pubchannel=pubchannel,
            host=host, serialization=msgpack_serialization)


class CustomException(Exception):
    pass


class Service(object):
    """
    Testing service that is register via the TestExecutor
    on the RPCServer
    """

    def __init__(self):
        self.data = {'foo': 'bar'}

    def multiply(self, x, y=1):
        return x * y

    def get_item(self, key):
        return self.data[key]

    def custom_error(self):
        raise CustomException("Foobar")


class ServiceClient(object):
    """
    TestService client, exposing (rpc) functions
    that can be called on the TestService instance.
    """
    def __init__(self, client):

        self.client = client

    async def multiply(self, x, y=100):
        rpc_func_call = RPCCall('multiply', [x], {'y': y})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)

    async def get_item(self, key):
        rpc_func_call = RPCCall('get_item', [key], {})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)

    async def custom_error(self):
        rpc_func_call = RPCCall('custom_error', [], {})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)


async def stop_rpc_server_on_result_of(
        async_func, rpc_server, rpc_client, client_processing=False):
    """
    awaits the given async_func.
    stops the rpc_server (background) processing on result,
    allowing the rpc_server.serve() to return.
    """
    # Await func result

    try:
        result = await async_func
    finally:
        # Stop listening and queue processing in server,
        # allowing rpc_serve.serve() to return
        await rpc_server.queue.put(b'END')
        await rpc_server.rpc_commlayer.unsubscribe()

        if client_processing:
            await rpc_client.queue.put(b'END')
            await rpc_client.rpc_commlayer.unsubscribe()

    # Return the result
    return result


@pytest.fixture
async def rpc_client():
    return RPCClient(await rpc_commlayer(b'pub', b'sub'))


@pytest.fixture
async def rpc_server():
    return RPCServer(await rpc_commlayer(b'sub', b'pub'))


@pytest.fixture
async def do_rpc_call():
    async def wrapper(service_client, executor, func, custom_dataclasses=[],
                      client_processing=False):
        # Initialize both client & server
        rpc_client = RPCClient(await rpc_commlayer(b'pub', b'sub'))
        rpc_server = RPCServer(await rpc_commlayer(b'sub', b'pub'))

        service_client.client = rpc_client
        rpc_server.register(executor)

        # Register any given custom dataclasses on both ends
        rpc_client.register_models(custom_dataclasses)
        rpc_server.register_models(custom_dataclasses)

        # Already start subscribing, to be sure
        # we don't miss a message
        await rpc_server.rpc_commlayer.do_subscribe()

        # Execute rpc call and stop rpc_server when
        # a result has been returned
        async_funcs = [
            stop_rpc_server_on_result_of(
                func, rpc_server, rpc_client, client_processing),
            rpc_server.serve()
        ]

        # Add rpc_client.serve if client processing
        if client_processing:
            async_funcs.append(rpc_client.serve())

        try:
            result = (await asyncio.gather(*async_funcs))[0]
        finally:
            # Close all rpc_commlayers
            await rpc_client.rpc_commlayer.close()
            await rpc_server.rpc_commlayer.close()

        return result

    return wrapper



import dataclasses
import msgpack
import numpy as np
from abc import ABC, abstractmethod
from io import BytesIO
from datetime import datetime
from lz4.frame import compress as lz4_compress, decompress as lz4_decompress
from typing import Any

# Maximum byte lengths for str/ext
MAX_STR_LEN = 2147483647
MAX_EXT_LEN = 2147483647


# Internal registry
# TODO: figure out if it is ok to do
# this on the module...
REGISTRY = {'obj_types': {},
            'ext_types': {},
            'serializables': {}}


def register(obj_def):
    """
    Register dataclasses or custom handlers in the registry.

    For example obj_def and required methods, see NumpyArray below
    """
    if dataclasses.is_dataclass(obj_def):
        # Handle dataclasses, every dataclass needs to be registered
        # via register.
        class_name = obj_def.__name__
        REGISTRY['serializables'][class_name] = obj_def
        REGISTRY['obj_types'][obj_def] = DataclassHandler

        # Register the DataclassHandler if not done already
        if DataclassHandler.ext_type not in REGISTRY['ext_types']:
            REGISTRY['ext_types'][
                DataclassHandler.ext_type] = DataclassHandler
    else:
        # Assume the obj_def has obj_type and ext_type, as can be
        # seen below.
        assert hasattr(obj_def, 'obj_type') and hasattr(obj_def, 'ext_type')
        REGISTRY['obj_types'][obj_def.obj_type] = obj_def
        REGISTRY['ext_types'][obj_def.ext_type] = obj_def


class AbstractHandler(ABC):
    ext_type: int = None  # Unique number
    obj_type: Any = None  # Unique object type

    @classmethod
    @abstractmethod
    def packb(cls, instance: Any) -> bytes:
        """
        Pack the instance into bytes
        """

    @classmethod
    @abstractmethod
    def unpackb(cls, data: bytes) -> Any:
        """
        Unpack the data back into an instance
        """


class NumpyArrayHandler(AbstractHandler):
    """
    Use np.save and np.load to serialize/deserialize
    numpy array's.
    """
    ext_type = 1
    obj_type = np.ndarray

    # Note:
    # More generic approach, but a bit slower than
    # packing it as a list/tuple with (dtype, shape, bytes)
    @classmethod
    def packb(cls, array: np.ndarray) -> bytes:
        buf = BytesIO()
        np.save(buf, array)
        buf.seek(0)
        return buf.read()

    @classmethod
    def unpackb(cls, data: bytes) -> np.ndarray:
        buf = BytesIO(data)
        buf.seek(0)
        return np.load(buf)


class NumpyStructuredArrayHandler(NumpyArrayHandler):
    ext_type = 2
    obj_type = np.void  # = the type of structured array's...


class DatetimeHandler:
    """
    Serialize datetime instances as timestamps.
    """
    ext_type = 3
    obj_type = datetime

    @classmethod
    def packb(cls, dt: datetime) -> bytes:
        return b'%f' % dt.timestamp()

    @classmethod
    def unpackb(cls, data: bytes) -> datetime:
        return datetime.fromtimestamp(float(data))


class DataclassHandler:
    """
    Serialize dataclasses by serializing the .__dict__
    of dataclasses. This allows recursively serialization for example:
    dataclasses in dataclasses or Numpy array's in dataclasses.
    """
    ext_type = 4

    @classmethod
    def packb(cls, obj) -> bytes:
        dataclass_name = obj.__class__.__name__
        if isinstance(dataclass_name, str):
            dataclass_name = dataclass_name

        # Recursively process dataclasses of the dataclass,
        # serialize as tuple(dataclass_name, __dict__)
        return dumpb(
            (dataclass_name, obj.__dict__),
            do_compress=False)

    @classmethod
    def unpackb(cls, data):
        # Recursively process the contents of the dataclass
        classname, data = loadb(
            data, do_decompress=False, raw=False)
        # Return registered class or Serializable (as default)
        assert classname in REGISTRY['serializables'], \
            f'class {classname} not yet registered'
        klass = REGISTRY['serializables'][classname]
        return klass(**data)


class SliceHandler:
    """
    Serialize slices
    """
    ext_type = 5
    obj_type = slice

    @classmethod
    def packb(cls, obj) -> bytes:
        return dumpb((obj.start, obj.stop, obj.step))

    @classmethod
    def unpackb(cls, data):
        return slice(*loadb(data))


# Register custom handlers
register(NumpyArrayHandler)
register(NumpyStructuredArrayHandler)
register(DatetimeHandler)
register(SliceHandler)


def default(obj: Any):
    """
    Serialize (dumpb) hook for obj types that msgpack does not
    process out of the box.
    """
    if type(obj) in REGISTRY['obj_types']:
        # If the type is in the registry, use the
        # handler to serialize the obj
        handler = REGISTRY['obj_types'][type(obj)]
        return msgpack.ExtType(
            handler.ext_type, handler.packb(obj))

    raise TypeError("Unknown type: %r" % (obj,))


def ext_hook(ext_type: int, bytes_data: bytes):
    """
    Deserialize (loadb) hook for ext_types that are
    not default in msgpack.

    ext_types are user defined numbers for special
    deserialization handling.
    """
    if ext_type in REGISTRY['ext_types']:
        # If the ext_type is in the registry, use the
        # handler to deserialize the bytes_data
        handler = REGISTRY['ext_types'][ext_type]
        return handler.unpackb(bytes_data)

    raise TypeError("Unknown ext_type: %r" % (ext_type,))  # pragma: no cover


def do_nothing(x):
    return x


def dumpb(instance: Any, do_compress=True, compress_func=lz4_compress,
          use_bin_type=True):
    """
    Dump/pack instance with msgpack to bytes
    """
    if not do_compress:
        compress_func = do_nothing
    return compress_func(msgpack.packb(
        instance, default=default, use_bin_type=use_bin_type))


def loadb(packed: bytes, do_decompress=True, decompress_func=lz4_decompress,
          raw=False):
    """
    Load/unpack bytes back to instance
    """
    if packed is None:
        return None
    if not do_decompress:
        decompress_func = do_nothing
    return msgpack.unpackb(
        decompress_func(packed), ext_hook=ext_hook,
        max_ext_len=MAX_EXT_LEN,
        max_str_len=MAX_STR_LEN, raw=raw)



class NotReceived(Exception):
    """
    Message has not been recieved by anyone
    """


class WrappedException(Exception):
    """
    Exception raised when an exception raised
    by RPC call could not be resolved, the innermessage
    shows the exception raised on the RPC server.
    """


class RPCTimeoutError(Exception):
    """
    Timeouterror raised by RPC client if result
    took to long.
    """


class SubscriptionClosed(Exception):
    """
    Raised when the subscription has already been closed
    """



import pytest
import numpy as np
import datetime
from dataclasses import dataclass
from asyncio_rpc.serialization import msgpack as msgpack_serialization


@pytest.fixture()
def serialize_deserialize():
    def func(value):
        return msgpack_serialization.loadb(
            msgpack_serialization.dumpb(value))
    return func


def test_dict_serialization(serialize_deserialize):
    value = {'1': 2, 10: 1.10}
    assert value == serialize_deserialize(value)


def test_byte_and_str_serialization(serialize_deserialize):
    value = {'1': b'2', b'10': '1.10'}
    assert value == serialize_deserialize(value)


def test_numpy_serialization(serialize_deserialize):
    value = np.arange(100, dtype=np.float64)
    assert np.all(value == serialize_deserialize(value))


def test_datetime_serialization(serialize_deserialize):
    value = datetime.datetime.now()
    assert value == serialize_deserialize(value)


def test_none_serialization(serialize_deserialize):
    assert serialize_deserialize(None) is None


def test_none_deserialization():
    assert msgpack_serialization.loadb(None) is None


def test_slice_serialization(serialize_deserialize):
    value = slice(1, 2, 3)
    assert value == serialize_deserialize(value)


@dataclass
class DataclassTest:
    uid: int
    data: np.ndarray


@dataclass
class DataclassWrapper:
    dataclass_test: DataclassTest


msgpack_serialization.register(DataclassTest)
msgpack_serialization.register(DataclassWrapper)


def test_dataclass_serialization(serialize_deserialize):
    value = DataclassTest(101, np.arange(100, dtype=np.float32))
    deserialized = serialize_deserialize(value)
    assert value.uid == deserialized.uid
    assert np.all(value.data == deserialized.data)


def test_dataclass_wrapper_serialization(serialize_deserialize):
    value = DataclassWrapper(
        DataclassTest(101, np.arange(100, dtype=np.float32)))
    deserialized = serialize_deserialize(value)
    assert value.dataclass_test.uid == deserialized.dataclass_test.uid
    assert np.all(
        value.dataclass_test.data == deserialized.dataclass_test.data)


@dataclass
class UnregisteredTest:
    uid: int
    data: np.ndarray


def test_unregistered_serialization(serialize_deserialize):
    value = UnregisteredTest(101, np.arange(100, dtype=np.float32))
    with pytest.raises(TypeError):
        serialize_deserialize(value)



import pytest
import asyncio
from uuid import uuid4
from asyncio_rpc.server import RPCServer
from asyncio_rpc.client import RPCClient

from asyncio_rpc.models import RPCCall, RPCSubStack
from .utils import rpc_commlayer
from asyncio_rpc.pubsub import Publisher


class Executor:
    """
    Default executor implementation, override if necessary
    """

    def __init__(self, namespace, instance):
        self.namespace = namespace
        self.instance = instance

    async def subscribe_call(self, publisher: Publisher):
        """
        Use the Publisher to publish results to the client
        """
        for i in range(0, 20):
            if publisher.is_active:
                await publisher.publish(i)
        for i in range(0, 20):
            if not publisher.is_active:
                break
            await publisher.publish(i)

        # Clean-up
        rpc_server = publisher._server
        await rpc_server.queue.put(b'END')
        await rpc_server.rpc_commlayer.unsubscribe()

    async def rpc_call(self, rpc_stack):
        pass


@pytest.mark.asyncio
async def test_pubsub():

    rpc_client = RPCClient(await rpc_commlayer(b'pub', b'sub'))
    rpc_server = RPCServer(await rpc_commlayer(b'sub', b'pub'))
    executor = Executor('PUBSUB', None)
    rpc_server.register(executor)

    await rpc_server.rpc_commlayer.do_subscribe()

    rpc_func_call = RPCCall('get_item', [1], {})
    rpc_func_stack = RPCSubStack(
        uuid4().hex, 'PUBSUB', 300, [rpc_func_call])

    async def process_subscriber(rpc_func_stack):
        subscriber = await rpc_client.subscribe_call(rpc_func_stack)

        async for item in subscriber.enumerate():
            if item > 5:
                await subscriber.close()

        # Clean-up
        await rpc_client.queue.put(b'END')
        await rpc_client.rpc_commlayer.unsubscribe()

    funcs = [
        rpc_server.serve(),
        rpc_client.serve(),
        process_subscriber(rpc_func_stack),
    ]
    await asyncio.gather(*funcs)

    await rpc_client.rpc_commlayer.close()
    await rpc_server.rpc_commlayer.close()



import asyncio
import logging
from typing import List
from asyncio_rpc.models import (
    RPCStack, RPCResult, RPCException, RPCCall,
    RPCSubStack, RPCUnSubStack)
from asyncio_rpc.commlayers.base import AbstractRPCCommLayer
from asyncio_rpc.pubsub import Publisher
logger = logging.getLogger("asyncio-rpc-server")


class NamespaceError(Exception):
    """
    Exception raised when a namespace unknown or
    already present.
    """


class RPCServer(object):
    """
    Remote procedure server class. Allows to register executors
    by namespace and execute RPC calls from a RPCClient
    """
    def __init__(self, rpc_commlayer: AbstractRPCCommLayer = None):
        assert isinstance(rpc_commlayer, AbstractRPCCommLayer)
        self.queue = asyncio.Queue()
        self._alive = True

        # Allow multiple executors to be registered by
        # namespace
        self.registry = {}
        self.publishers = {}
        self.rpc_commlayer = rpc_commlayer

    def register_models(self, models):
        """
        Register all given models to the rpc_commlayer serialization

        Intended usage is to register dataclasses
        """
        for model in models:
            self.rpc_commlayer.serialization.register(model)

    def register(self, executor):
        """
        Register an executor for a namespace, the namespace
        should be unique and is used to route RPC calls from
        the client to the correct executor in the registry of
        the RPCServer.
        """
        assert hasattr(executor, 'namespace')

        if executor.namespace in self.registry:
            raise NamespaceError("Namespace already exists")

        # Register executor for this namespaces
        self.registry[executor.namespace] = executor

    async def rpc_call(self, rpc_func_stack: RPCStack):
        """
        Incoming rpc_call, execute it via the registered
        executor and return the result or exception
        """
        assert isinstance(rpc_func_stack, RPCStack)

        # Create a default result
        result = RPCResult(
            uid=rpc_func_stack.uid,
            namespace=rpc_func_stack.namespace,
            data=None)

        if rpc_func_stack.namespace not in self.registry:
            raise NamespaceError("Unknown namespace")

        executor = self.registry[rpc_func_stack.namespace]

        try:
            # Wait for result from executor
            result.data = await asyncio.wait_for(
                executor.rpc_call(rpc_func_stack.stack),
                timeout=rpc_func_stack.timeout)
        except Exception as e:
            # For now catch all exceptions here...
            # TODO: debug mode with stacktrace
            result = RPCException(
                uid=rpc_func_stack.uid,
                namespace=rpc_func_stack.namespace,
                classname=e.__class__.__name__,
                exc_args=e.args)

        return result

    async def subscribe_call(self, rpc_sub_stack: RPCSubStack):
        assert isinstance(rpc_sub_stack, RPCSubStack)
        if rpc_sub_stack.namespace not in self.registry:
            raise NamespaceError("Unknown namespace")

        executor = self.registry[rpc_sub_stack.namespace]
        if not hasattr(executor, 'subscribe_call'):
            raise NotImplementedError(
                f"Executor for namespace: {rpc_sub_stack.namespace} has "
                f"no subscribe_call function")

        publisher = Publisher(self, rpc_sub_stack)
        self.publishers[rpc_sub_stack.uid] = publisher

        # Create task for this publisher
        asyncio.create_task(
            executor.subscribe_call(publisher))

    async def _on_rpc_event(
            self, rpc_func_stack: RPCStack, channel: bytes = None):
        """
        Callback function sent to rpc_commlayer, is called
        when a RPCStack is received by the rpc_commlayer subscription
        """
        await self.queue.put((rpc_func_stack, channel))

    async def _process_queue(self):
        """
        Background queue processing function, processes
        the internal self.queue until b'END' is received


        if on_rpc_message has been set, it will be called
        whenever a RPCMessage is popped from the queue.
        """

        while self._alive:
            item = await self.queue.get()

            if item == b'END':
                break

            rpc_func_stack, channel = item

            assert isinstance(rpc_func_stack, RPCStack)

            if isinstance(rpc_func_stack, RPCSubStack):
                try:
                    # Process rpc_func_call_stack
                    await self.subscribe_call(rpc_func_stack)
                except Exception as e:
                    result = RPCException(
                        uid=rpc_func_stack.uid,
                        namespace=rpc_func_stack.namespace,
                        classname=e.__class__.__name__,
                        exc_args=e.args)
                    # Publish exception
                    await self.rpc_commlayer.publish(
                        result, channel=rpc_func_stack.respond_to)
            elif isinstance(rpc_func_stack, RPCUnSubStack):
                publisher = self.publishers.pop(rpc_func_stack.uid, None)
                if publisher is not None:
                    publisher.set_is_active(False)
            else:
                try:
                    # Process rpc_func_call_stack
                    result = await self.rpc_call(rpc_func_stack)
                    # Publish result of rpc call
                    await self.rpc_commlayer.publish(
                        result, channel=rpc_func_stack.respond_to)

                except Exception as e:
                    result = RPCException(
                        uid=rpc_func_stack.uid,
                        namespace=rpc_func_stack.namespace,
                        classname=e.__class__.__name__,
                        exc_args=e.args)
                    # Try to publish error
                    await self.rpc_commlayer.publish(
                        result, channel=rpc_func_stack.respond_to)

    async def serve(self):
        """
        Main entry point for RPCServer.

        Starts RPCServer background processing, blocks
        until self.rpc_commlayer.unsubscribe() is called.
        """
        task_args_map = {
            self.rpc_commlayer.subscribe: [self._on_rpc_event],
            self._process_queue: []
        }

        # create the main tasks
        main_tasks = {
            asyncio.ensure_future(coro(*args)): (coro, args)
            for coro, args in task_args_map.items()
        }

        running = set(main_tasks.keys())

        except_cnt = -1

        while running:
            except_cnt += 1
            finished, running = await asyncio.wait(
                running, return_when=asyncio.FIRST_EXCEPTION
            )
            for task in finished:
                if task.exception():
                    logger.exception(task.exception())
                    task.print_stack()
                    coro, args = main_tasks[task]
                    new_task = asyncio.ensure_future(coro(*args))
                    main_tasks[new_task] = (coro, args)
                    running.add(new_task)


class DefaultExecutor:
    """
    Default executor implementation, override if necessary
    """

    def __init__(self, namespace, instance):
        assert namespace is not None
        assert instance is not None
        self.namespace = namespace
        self.instance = instance

    async def subscribe_call(self, publisher: Publisher):
        """
        Use the Publisher to publish results to the client
        """
        # if publisher.is_active:
        #     await publisher.publish(b'blaat')
        pass

    async def rpc_call(self, stack: List[RPCCall] = []):
        """
        Process incoming rpc call stack.
        The stack can contain multiple chained function calls for example:
            node.filter(id=1).reproject_to('4326').data
        """

        resource = self.instance

        for rpc_func_call in stack:
            assert isinstance(rpc_func_call, RPCCall)

            # Try to get the function/property from self.instance
            instance_attr = getattr(resource, rpc_func_call.func_name)

            if callable(instance_attr):
                # Function
                resource = instance_attr(
                    *rpc_func_call.func_args,
                    **rpc_func_call.func_kwargs)
            else:
                # Asume property
                resource = instance_attr

        return resource



import pytest

from tests.utils import (
    rpc_server as rpc_server_fixture
)
from asyncio_rpc.server import NamespaceError, DefaultExecutor
from asyncio_rpc.models import RPCStack

rpc_server = rpc_server_fixture


class MockService:
    pass


@pytest.mark.asyncio
async def test_registration(rpc_server):
    rpc_server.register(DefaultExecutor("TEST", MockService()))
    await rpc_server.rpc_commlayer.close()


@pytest.mark.asyncio
async def test_double_registration_error(rpc_server):
    rpc_server.register(DefaultExecutor("TEST", MockService()))

    with pytest.raises(NamespaceError):
        rpc_server.register(DefaultExecutor("TEST", MockService()))

    await rpc_server.rpc_commlayer.close()


@pytest.mark.asyncio
async def test_unknown_namespace_error(rpc_server):
    with pytest.raises(NamespaceError):
        await rpc_server.rpc_call(
            RPCStack(uid='1', namespace='UNKNOWN', stack=[], timeout=300))

    await rpc_server.rpc_commlayer.close()



from dataclasses import dataclass
from typing import List, Dict, Any


class RPCBase:
    """
    Baseclass to identify all RPC dataclasses
    """


@dataclass
class RPCMessage(RPCBase):
    """
    Message definition. Can be used to publish some data

    :param uid: the unique id for this RPC message
    :param namespace: the namespace
    :param data: the data to sent (should not be to much data!)
    """
    uid: str
    namespace: str
    data: Any


@dataclass
class RPCCall(RPCBase):
    """
    Remote procedure call definition

    :param name: the name of the function to execute
    :param args: args for the function
    :param kwargs: kwargs for the function
    """
    func_name: str
    func_args: List
    func_kwargs: Dict


@dataclass
class RPCStack(RPCBase):
    """
    Represents a remote Procedure function call stack, for example the
    call:
        gr.nodes.subset('2D_open_water').filter(id__lt=100).reproject_to('4326').coordinates

    Would procedure the following stack:
        [
            RPCCall('nodes', [], {}),
            RPCCall('subset',['2D_open_water'], {}),
            RPCCall('filter',[], {'id__lt': 100}),
            RPCCall('reproject_to',['4326'], {}),
            RPCCall('coordinates', [], {}),
        ]

    Note: properties are also encoded as function calls.
    """
    uid: str
    namespace: str
    timeout: float
    stack: List[RPCCall]
    respond_to: str = None


@dataclass
class RPCSubStack(RPCStack):
    """
    Same as RPCStack, but no for subscribing to a generator
    """


@dataclass
class RPCUnSubStack(RPCStack):
    """
    Unsubscribe call
    """


@dataclass
class RPCResult(RPCBase):
    """
    Represents the result of a remote procedure call (RPCStack)

    :param uid: is set to RPCStack.uid
    :param namespace: is set to RPCStack.namespace
    :param data: is the result of the RPCStack call
    """
    uid: str
    namespace: str
    data: Any


@dataclass
class RPCPubResult(RPCResult):
    """
    Same as RPCResult, but now for publications
    """


@dataclass
class RPCException(RPCBase):
    """
    Represents an exception raised during executing the remote procedure
    call server-side

    :param uid: is set to RPCStack.uid
    :param namespace: is set to RPCStack.namespace
    :param classname: the classname of the exception raised
    :param exc_args: the exception args as a list
    """
    uid: str
    namespace: str
    classname: str # noqa
    exc_args: List


# List of serializable model to
# register in the serialization
SERIALIZABLE_MODELS = (
    RPCMessage,
    RPCCall,
    RPCStack,
    RPCSubStack,
    RPCUnSubStack,
    RPCResult,
    RPCPubResult,
    RPCException,
)



import pytest
from dataclasses import dataclass
from uuid import uuid4


from tests.utils import (
    do_rpc_call as do_rpc_call_fixture
)

from asyncio_rpc.models import RPCCall, RPCStack
from asyncio_rpc.client import WrappedException
from asyncio_rpc.server import DefaultExecutor


do_rpc_call = do_rpc_call_fixture


@dataclass
class CustomDataModel:
    x: int
    y: int

    def multiply(self):
        return self.x * self.y


class CustomException(Exception):
    pass


class Service(object):
    """
    Testing service that is register via the TestExecutor
    on the RPCServer
    """

    def __init__(self):
        self._data = {'foo': 'bar'}

    def multiply(self, x, y=1):
        return x * y

    @property
    def data(self):
        return self._data

    def get_item(self, key):
        return self._data[key]

    def custom_error(self):
        raise CustomException("Foobar")

    def multiply_with_dataclass(self, x: CustomDataModel):
        assert isinstance(x, CustomDataModel)
        return x.multiply()


class ServiceClient(object):
    """
    TestService client, exposing (rpc) functions
    that can be called on the TestService instance.
    """
    def __init__(self, client):
        self.client = client

    @property
    async def data(self):
        rpc_func_call = RPCCall('data', [], {})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)

    async def multiply(self, x, y=100):
        rpc_func_call = RPCCall('multiply', [x], {'y': y})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)

    async def get_item(self, key):
        rpc_func_call = RPCCall('get_item', [key], {})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)

    async def custom_error(self):
        rpc_func_call = RPCCall('custom_error', [], {})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)

    async def multiply_with_dataclass(self, x: CustomDataModel):
        assert isinstance(x, CustomDataModel)
        rpc_func_call = RPCCall('multiply_with_dataclass', [x], {})
        rpc_func_stack = RPCStack(
            uuid4().hex, 'TEST', 300, [rpc_func_call])
        return await self.client.rpc_call(rpc_func_stack)


@pytest.mark.asyncio
async def test_simple_call(do_rpc_call):
    test_service_client = ServiceClient(None)
    result = await do_rpc_call(
        test_service_client,
        DefaultExecutor("TEST", Service()),
        test_service_client.multiply(100, 100))
    assert result == 100 * 100


@pytest.mark.asyncio
async def test_simple_call_with_client_processing(do_rpc_call):
    test_service_client = ServiceClient(None)
    result = await do_rpc_call(
        test_service_client,
        DefaultExecutor("TEST", Service()),
        test_service_client.multiply(100, 100),
        client_processing=True)
    assert result == 100 * 100


@pytest.mark.asyncio
async def test_simple_call2(do_rpc_call):
    test_service_client = ServiceClient(None)
    result = await do_rpc_call(
        test_service_client,
        DefaultExecutor("TEST", Service()),
        test_service_client.get_item('foo'))
    assert result == 'bar'


@pytest.mark.asyncio
async def test_property(do_rpc_call):
    test_service_client = ServiceClient(None)
    result = await do_rpc_call(
        test_service_client,
        DefaultExecutor("TEST", Service()),
        test_service_client.data)
    assert result == {'foo': 'bar'}


@pytest.mark.asyncio
async def test_key_error(do_rpc_call):
    test_service_client = ServiceClient(None)
    with pytest.raises(KeyError):
        await do_rpc_call(
            test_service_client,
            DefaultExecutor("TEST", Service()),
            test_service_client.get_item('bar'))


@pytest.mark.asyncio
async def test_not_builtin_exception(do_rpc_call):
    test_service_client = ServiceClient(None)
    with pytest.raises(WrappedException):
        await do_rpc_call(
            test_service_client,
            DefaultExecutor("TEST", Service()),
            test_service_client.custom_error())


@pytest.mark.asyncio
async def test_custom_data_model(do_rpc_call):
    test_service_client = ServiceClient(None)
    value = CustomDataModel(100, 100)
    result = await do_rpc_call(
            test_service_client,
            DefaultExecutor("TEST", Service()),
            test_service_client.multiply_with_dataclass(value),
            custom_dataclasses=[CustomDataModel])
    assert result == value.multiply()



import asyncio
import builtins
import logging
from typing import Union, List
from .models import (
    RPCMessage, RPCResult, RPCException, RPCStack, RPCBase, RPCSubStack)
from asyncio_rpc.commlayers.base import AbstractRPCCommLayer
from asyncio_rpc.pubsub import Subscription
from asyncio_rpc.exceptions import (
    RPCTimeoutError, WrappedException, NotReceived)

logger = logging.getLogger("asyncio-rpc-client")


class RPCClient(object):
    """
    Remote procedure client class. Allows to send rpc_call
    to a RPCServer via a rpc_commlayer.
    """

    def __init__(self, rpc_commlayer: AbstractRPCCommLayer):
        """
        Initialize a new RPCClient by providing an implementation of
        AbstractRPCCommlayer
        """
        assert isinstance(rpc_commlayer, AbstractRPCCommLayer)
        self.rpc_commlayer = rpc_commlayer
        self.futures = {}
        self.queue = asyncio.Queue()
        self.processing = False
        self.on_rpc_message = None
        self.subscriptions = {}

    def register_models(self, models: List):
        """
        Register all given models to the rpc_commlayer serialization

        Intended usage is to register dataclasses
        """
        for model in models:
            self.rpc_commlayer.serialization.register(model)

    async def _wait_for_result(
            self, uid: bytes) -> Union[RPCResult, RPCException]:
        """
        Internal helper function for stopping the rpc_commlayer subscription
        upon receiving a result from the RPC_server
        """
        result = None
        while True:
            event, channel = await self.queue.get()

            assert isinstance(event, RPCResult) or\
                isinstance(event, RPCException)

            # Discard everything that we don't need...
            if event.uid == uid:
                result = event
                break

        # Stop subscription, automatically stops
        # queue processing as well
        await self.rpc_commlayer.unsubscribe()

        return result

    async def subscribe_call(
            self, rpc_sub_stack: RPCSubStack, channel=None) -> Subscription:
        assert isinstance(rpc_sub_stack, RPCStack)

        if not self.processing:
            pass

        # Make sure to be subscribed before publishing
        await self.rpc_commlayer.do_subscribe()

        subscription = Subscription(self, rpc_sub_stack)

        self.subscriptions[rpc_sub_stack.uid] = subscription

        # Publish RPCStack to RPCServer
        count = await self.rpc_commlayer.publish(
            rpc_sub_stack, channel=channel)

        if count == 0:
            raise NotReceived(
                f"subscribe_call was not "
                f"received by any server: {rpc_sub_stack}")

        return subscription

    async def rpc_call(
            self, rpc_func_stack: RPCStack, channel=None) -> RPCResult:
        """
        Execute the given rpc_func_stack (RPCStack) and either
        return a RPCResult or raise an exception based on the returned
        RPCException.

        This function can both be called with awaiting client.serve() or
        without. The difference is, that in the first case client.serve()
        starts while loops for performing background processing.

        The channel (optional) argument can be used to override
        the default publish channel
        """
        assert isinstance(rpc_func_stack, RPCStack)

        # Make sure to be subscribed before publishing
        await self.rpc_commlayer.do_subscribe()

        # Publish RPCStack to RPCServer
        count = await self.rpc_commlayer.publish(
            rpc_func_stack, channel=channel)

        if count == 0:
            raise NotReceived(
                f"rpc_call was not received "
                f"by any subscriber {rpc_func_stack}")

        if self.processing:
            # client.serve() has been awaited, so
            # background processing is taken care of.

            # Create a future that should be resolved within
            # the given timeout. The background processing initialized
            # by client.serve() resolves the future when a result
            # is returned.
            future = asyncio.get_event_loop().create_future()
            self.futures[rpc_func_stack.uid] = future
            try:
                result = await asyncio.wait_for(
                    future, timeout=rpc_func_stack.timeout)
            except asyncio.TimeoutError:
                raise RPCTimeoutError(f"rpc_func_stack: {rpc_func_stack}")
        else:
            # No background processing, so start the subscription
            # and wait for result via asyncio.gather
            # _wait_for_result is a helper function to unsubscribe
            # on a result.
            try:
                _, result = await asyncio.gather(
                    self.rpc_commlayer.subscribe(self._on_rpc_event),
                    self._wait_for_result(rpc_func_stack.uid)
                )
            except asyncio.TimeoutError:
                raise RPCTimeoutError(f"rpc_func_stack: {rpc_func_stack}")

        if isinstance(result, RPCException):
            # Try to resolve builtin errors
            try:
                exception_class = getattr(builtins, result.classname)
            except AttributeError:
                # Default to WrappedException if
                # returned exception is not a builtin error
                exception_class = WrappedException

            raise exception_class(*result.exc_args)

        return result.data

    async def _on_rpc_event(
            self, rpc_instance: RPCBase, channel: bytes = None):
        """
        Callback function sent to rpc_commlayer, is called
        when a message is received by the subscription.
        """
        assert isinstance(rpc_instance, RPCBase)
        # Put everything in a queue and process
        # it afterwards
        await self.queue.put((rpc_instance, channel))

    async def _process_queue(self, on_rpc_message: callable = None):
        """
        Background queue processing function, processes
        the internal self.queue until b'END' is received


        if on_rpc_message has been set, it will be called
        whenever a RPCMessage is popped from the queue.
        """
        self.processing = True

        while True:
            item = await self.queue.get()
            if item == b'END':
                break

            event, channel = item
            # The event can either be a:
            # 1) RPCResult or RPCException as a result from the RPCServer
            # 3) RPCMessage as a message from the RPCServer
            assert isinstance(event, RPCResult) or\
                isinstance(event, RPCException) or\
                isinstance(event, RPCMessage)

            if isinstance(event, RPCResult) or isinstance(event, RPCException):
                if event.uid in self.futures:
                    # A future is created & awaited in self.rpc_call
                    # resolve this future to proceed in the rpc_call function
                    # and return a result
                    future = self.futures.pop(event.uid)
                    if future is not None:
                        future.set_result(event)
                elif event.uid in self.subscriptions:
                    subscription = self.subscriptions[event.uid]
                    await subscription.enqueue(event)
                else:
                    # FUTURE NOT FOUND FOR EVENT
                    # TODO: how to handle this??
                    pass  # pragma: nocover
            elif isinstance(event, RPCMessage) and on_rpc_message:
                await on_rpc_message(event, channel)

        self.processing = False

    async def serve(self, on_rpc_message=None):
        """
        Start RPCClient background processing, blocks
        until self.rpc_commlayer.unsubscribe() is called.

        Use this method in an async context to enable
        background processing and allowing multiple
        rpc_calls asynchroniously.
        """

        task_args_map = {
            self.rpc_commlayer.subscribe: [self._on_rpc_event],
            self._process_queue: [on_rpc_message]
        }
        # create the main tasks
        main_tasks = {
            asyncio.ensure_future(coro(*args)): (coro, args)
            for coro, args in task_args_map.items()
        }

        running = set(main_tasks.keys())

        except_cnt = -1

        while running:
            except_cnt += 1
            finished, running = await asyncio.wait(
                running, return_when=asyncio.FIRST_EXCEPTION
            )
            for task in finished:
                if task.exception():
                    logger.exception(task.exception())
                    task.print_stack()
                    coro, args = main_tasks[task]
                    new_task = asyncio.ensure_future(coro(*args))
                    main_tasks[new_task] = (coro, args)
                    running.add(new_task)



from uuid import uuid4
from .base import AbstractRPCCommLayer
from aioredis import create_redis
from ..models import RPCStack, RPCResult, RPCBase, SERIALIZABLE_MODELS

RESULT_EXPIRE_TIME = 300  # seconds


class RPCRedisCommLayer(AbstractRPCCommLayer):
    """
    Redis remote procedure call communication layer
    """

    @classmethod
    async def create(
            cls, subchannel=b'subchannel', pubchannel=b'pubchannel',
            host='localhost', port=6379, serialization=None):
        """
        Use a static create method to allow async context,
        __init__ cannot be async.
        """

        self = RPCRedisCommLayer(subchannel, pubchannel)

        # Create communicationLayer
        self.host = host
        self.port = port
        self.serialization = serialization

        # Redis for publishing
        self.redis = await create_redis(
            f'redis://{host}')

        # By default register all RPC models
        for model in SERIALIZABLE_MODELS:
            # Register models to serialization
            serialization.register(model)

        self.subscribed = False

        # Subscription has own redis
        self.sub_redis = None
        self.sub_channel = None

        # By default subscribe
        await self.do_subscribe()

        return self

    def __init__(self, subchannel, pubchannel):
        """
        Initialize and set the sub/pub channels
        """
        self.subchannel = subchannel
        self.pubchannel = pubchannel

    async def do_subscribe(self):
        if not self.subscribed:
            # By default subscribe
            self.sub_redis = await create_redis(
                f'redis://{self.host}')
            channels = await self.sub_redis.subscribe(
                self.subchannel)
            self.sub_channel = channels[0]
            self.subscribed = True

    async def publish(self, rpc_instance: RPCBase, channel=None):
        """
        Publish redis implementation, publishes RPCBase instances.

        :return: the number of receivers
        """
        # rpc_instance should be a subclass of RPCBase
        # For now just check if instance of RPCBase
        assert isinstance(rpc_instance, RPCBase)

        if isinstance(rpc_instance, RPCStack):
            # Add subchannel to RPCStack as respond_to
            rpc_instance.respond_to = self.subchannel
        elif ((isinstance(rpc_instance, RPCResult) and
               rpc_instance.data is not None)):
            # Customized:
            # result data via redis.set
            # result without data via redis.publish
            redis_key = uuid4().hex

            # Store the result data via key/value in redis
            await self.redis.set(
                redis_key,
                self.serialization.dumpb(rpc_instance.data),
                expire=RESULT_EXPIRE_TIME)

            # Set redis_key and remove data, since
            # this is stored in redis now
            rpc_instance.data = {'redis_key': redis_key}

        # Override the pub_channel with channel, if set
        pub_channel = channel if channel is not None else self.pubchannel

        # Publish rpc_instance and return number of listeners
        return await self.redis.publish(
            pub_channel,
            self.serialization.dumpb(rpc_instance))

    async def get_data(self, redis_key, delete=True):
        """
        Helper function to get data by redis_key, by default
        delete the data after retrieval.
        """
        data = self.serialization.loadb(
            await self.redis.get(redis_key))
        if delete:
            await self.redis.delete(redis_key)
        return data

    async def _process_msg(self, msg, on_rpc_event_callback, channel=None):
        """
        Interal message processing, is called on every received
        message via the subscription.
        """
        event = self.serialization.loadb(msg)

        # rpc_instance should be a subclass of RPCBase
        # For now just check if instance of RPCBase
        assert isinstance(event, RPCBase)

        if on_rpc_event_callback:
            if isinstance(event, RPCResult):
                # Customized:
                # result data via redis.set
                # result without data via redis.publish

                # Get data from redis and put it on the event
                if isinstance(event.data, dict) and 'redis_key' in event.data:
                    event.data = await self.get_data(event.data['redis_key'])

            await on_rpc_event_callback(
                event, channel=channel.name)

    async def subscribe(self, on_rpc_event_callback, channel=None, redis=None):
        """
        Redis implementation for subscribe method, receives messages from
        subscription channel.

        Note: does block in while loop until .unsubscribe() is called.
        """
        try:
            if channel is None:
                channel = self.sub_channel
            if redis is None:
                redis = self.sub_redis

            self.subscribed = True
            # Inside a while loop, wait for incoming events.
            while await channel.wait_message():
                await self._process_msg(
                    await channel.get(),
                    on_rpc_event_callback,
                    channel=channel)

        finally:
            # Close connections and cleanup
            self.subscribed = False
            redis.close()
            await redis.wait_closed()

    async def unsubscribe(self):
        """
        Redis implementation for unsubscribe. Stops subscription and breaks
        out of the while loop in .subscribe()
        """
        if self.subscribed:
            await self.sub_redis.unsubscribe(
                self.sub_channel.name)
            self.subscribed = False

    async def close(self):
        """
        Stop subscription & close everything
        """
        await self.unsubscribe()
        self.redis.close()
        await self.redis.wait_closed()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

import codecs
import re
import os

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

lightweight_requirements = [
    'numpy>=1.13',
    'msgpack>=0.6.0',
    'lz4>=2.1.6',
    'aioredis>=1.2.0'
]

setup_requirements = []

test_requirements = ['pytest==4.0.1']

setup(
    author="Jelle Prins",
    author_email='jelle.prins@nelen-schuurmans.nl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
    ],
    description="Asyncio RPC client/server with redis/msgpack/dataclasses",
    entry_points={
    },
    install_requires=lightweight_requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='asyncio rpc',
    name='asyncio_rpc',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests",
                 "__pycache__", "*.__pycache__.*"]),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    extras_require={
    },
    url='https://github.com/nens/asyncio_rpc',
    version=find_version("asyncio_rpc", "__init__.py"),
    zip_safe=False,
)



import asyncio
import builtins
from typing import Any, AsyncIterator
from asyncio_rpc.models import (
    RPCStack, RPCPubResult, RPCException, RPCUnSubStack)
from asyncio_rpc.exceptions import WrappedException


class Publisher:
    def __init__(self, server, rpc_stack: RPCStack):
        self._rpc_stack = rpc_stack
        self._server = server
        self._is_active = True

    def set_is_active(self, is_active: bool):
        self._is_active = is_active

    @property
    def is_active(self):
        return self._is_active

    @property
    def rpc_stack(self):
        return self._rpc_stack

    async def publish(self, data: Any):
        """
        Publish data to the client
        """
        if not self.is_active:
            return 0

        # Publish the data as partial data
        publication = RPCPubResult(
            self._rpc_stack.uid, self._rpc_stack.namespace, data)
        receiver_count = await self._server.rpc_commlayer.publish(
            publication, channel=self._rpc_stack.respond_to)

        if receiver_count == 0:
            self._is_active = False
            self._server.publishers.pop(self._rpc_stack.uid, None)
            return 0

        return receiver_count

    def __del__(self):
        self._server.publishers.pop(self._rpc_stack.uid, None)


class Subscription:
    def __init__(self, client, rpc_stack: RPCStack):
        self.queue = asyncio.Queue()
        self._client = client
        self._rpc_stack = rpc_stack

    async def enqueue(self, data: Any):
        await self.queue.put(data)

    async def close(self):
        self._client.subscriptions.pop(self._rpc_stack.uid, None)

        rpc_unsub_stack = RPCUnSubStack(
            self._rpc_stack.uid, self._rpc_stack.namespace,
            300, [None])

        # Publish to RPCServer
        await self._client.rpc_commlayer.publish(
            rpc_unsub_stack)

        self.queue._queue.clear()
        self.queue._finished.set()
        self.queue._unfinished_tasks = 0
        await self.queue.put(b'STOP')

    async def enumerate(self) -> AsyncIterator[Any]:
        while True:
            result = await self.queue.get()
            if result == b'STOP':
                break

            if isinstance(result, RPCException):
                # Try to resolve builtin errors
                try:
                    exception_class = getattr(builtins, result.classname)
                except AttributeError:
                    # Default to WrappedException if
                    # returned exception is not a builtin error
                    exception_class = WrappedException

                raise exception_class(*result.exc_args)

            yield result.data

    def __del__(self):
        self._client.subscriptions.pop(
            self._rpc_stack.uid, None)
